/*
 * JOCL - Java bindings for OpenCL
 * Copyright 2010 Marco Hutter - http://www.jocl.org/
 */

package edu.dhbw.mannheim.tigers.sumatra.model.data.math;

import static org.jocl.CL.CL_DEVICE_ADDRESS_BITS;
import static org.jocl.CL.CL_DEVICE_ERROR_CORRECTION_SUPPORT;
import static org.jocl.CL.CL_DEVICE_GLOBAL_MEM_SIZE;
import static org.jocl.CL.CL_DEVICE_IMAGE2D_MAX_HEIGHT;
import static org.jocl.CL.CL_DEVICE_IMAGE2D_MAX_WIDTH;
import static org.jocl.CL.CL_DEVICE_IMAGE3D_MAX_DEPTH;
import static org.jocl.CL.CL_DEVICE_IMAGE3D_MAX_HEIGHT;
import static org.jocl.CL.CL_DEVICE_IMAGE3D_MAX_WIDTH;
import static org.jocl.CL.CL_DEVICE_IMAGE_SUPPORT;
import static org.jocl.CL.CL_DEVICE_LOCAL_MEM_SIZE;
import static org.jocl.CL.CL_DEVICE_LOCAL_MEM_TYPE;
import static org.jocl.CL.CL_DEVICE_MAX_CLOCK_FREQUENCY;
import static org.jocl.CL.CL_DEVICE_MAX_COMPUTE_UNITS;
import static org.jocl.CL.CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE;
import static org.jocl.CL.CL_DEVICE_MAX_MEM_ALLOC_SIZE;
import static org.jocl.CL.CL_DEVICE_MAX_READ_IMAGE_ARGS;
import static org.jocl.CL.CL_DEVICE_MAX_WORK_GROUP_SIZE;
import static org.jocl.CL.CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS;
import static org.jocl.CL.CL_DEVICE_MAX_WORK_ITEM_SIZES;
import static org.jocl.CL.CL_DEVICE_MAX_WRITE_IMAGE_ARGS;
import static org.jocl.CL.CL_DEVICE_NAME;
import static org.jocl.CL.CL_DEVICE_PREFERRED_VECTOR_WIDTH_CHAR;
import static org.jocl.CL.CL_DEVICE_PREFERRED_VECTOR_WIDTH_DOUBLE;
import static org.jocl.CL.CL_DEVICE_PREFERRED_VECTOR_WIDTH_FLOAT;
import static org.jocl.CL.CL_DEVICE_PREFERRED_VECTOR_WIDTH_INT;
import static org.jocl.CL.CL_DEVICE_PREFERRED_VECTOR_WIDTH_LONG;
import static org.jocl.CL.CL_DEVICE_PREFERRED_VECTOR_WIDTH_SHORT;
import static org.jocl.CL.CL_DEVICE_QUEUE_PROPERTIES;
import static org.jocl.CL.CL_DEVICE_SINGLE_FP_CONFIG;
import static org.jocl.CL.CL_DEVICE_TYPE;
import static org.jocl.CL.CL_DEVICE_TYPE_ACCELERATOR;
import static org.jocl.CL.CL_DEVICE_TYPE_ALL;
import static org.jocl.CL.CL_DEVICE_TYPE_CPU;
import static org.jocl.CL.CL_DEVICE_TYPE_DEFAULT;
import static org.jocl.CL.CL_DEVICE_TYPE_GPU;
import static org.jocl.CL.CL_DEVICE_VENDOR;
import static org.jocl.CL.CL_DRIVER_VERSION;
import static org.jocl.CL.CL_PLATFORM_NAME;
import static org.jocl.CL.CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE;
import static org.jocl.CL.CL_QUEUE_PROFILING_ENABLE;
import static org.jocl.CL.clGetDeviceIDs;
import static org.jocl.CL.clGetDeviceInfo;
import static org.jocl.CL.clGetPlatformIDs;
import static org.jocl.CL.clGetPlatformInfo;
import static org.jocl.CL.stringFor_cl_device_fp_config;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.jocl.Pointer;
import org.jocl.Sizeof;
import org.jocl.cl_device_id;
import org.jocl.cl_platform_id;


/**
 * A JOCL program that queries and prints information about all
 * available devices.
 */
public class JOCLDeviceQuery
{
	/**
	 * The entry point of this program
	 * 
	 * @param args Not used
	 */
	public static void main(final String args[])
	{
		// Obtain the number of platforms
		int numPlatforms[] = new int[1];
		clGetPlatformIDs(0, null, numPlatforms);
		
		System.out.println("Number of platforms: " + numPlatforms[0]);
		
		// Obtain the platform IDs
		cl_platform_id platforms[] = new cl_platform_id[numPlatforms[0]];
		clGetPlatformIDs(platforms.length, platforms, null);
		
		// Collect all devices of all platforms
		List<cl_device_id> devices = new ArrayList<cl_device_id>();
		for (cl_platform_id platform : platforms)
		{
			String platformName = getString(platform, CL_PLATFORM_NAME);
			
			// Obtain the number of devices for the current platform
			int numDevices[] = new int[1];
			clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, 0, null, numDevices);
			
			System.out.println("Number of devices in platform " + platformName + ": " + numDevices[0]);
			
			cl_device_id devicesArray[] = new cl_device_id[numDevices[0]];
			clGetDeviceIDs(platform, CL_DEVICE_TYPE_ALL, numDevices[0], devicesArray, null);
			
			devices.addAll(Arrays.asList(devicesArray));
		}
		
		// Print the infos about all devices
		for (cl_device_id device : devices)
		{
			// CL_DEVICE_NAME
			String deviceName = getString(device, CL_DEVICE_NAME);
			System.out.println("--- Info for device " + deviceName + ": ---");
			System.out.printf("CL_DEVICE_NAME: \t\t\t%s%n", deviceName);
			
			// CL_DEVICE_VENDOR
			String deviceVendor = getString(device, CL_DEVICE_VENDOR);
			System.out.printf("CL_DEVICE_VENDOR: \t\t\t%s%n", deviceVendor);
			
			// CL_DRIVER_VERSION
			String driverVersion = getString(device, CL_DRIVER_VERSION);
			System.out.printf("CL_DRIVER_VERSION: \t\t\t%s%n", driverVersion);
			
			// CL_DEVICE_TYPE
			long deviceType = getLong(device, CL_DEVICE_TYPE);
			if ((deviceType & CL_DEVICE_TYPE_CPU) != 0)
			{
				System.out.printf("CL_DEVICE_TYPE:\t\t\t\t%s%n", "CL_DEVICE_TYPE_CPU");
			}
			if ((deviceType & CL_DEVICE_TYPE_GPU) != 0)
			{
				System.out.printf("CL_DEVICE_TYPE:\t\t\t\t%s%n", "CL_DEVICE_TYPE_GPU");
			}
			if ((deviceType & CL_DEVICE_TYPE_ACCELERATOR) != 0)
			{
				System.out.printf("CL_DEVICE_TYPE:\t\t\t\t%s%n", "CL_DEVICE_TYPE_ACCELERATOR");
			}
			if ((deviceType & CL_DEVICE_TYPE_DEFAULT) != 0)
			{
				System.out.printf("CL_DEVICE_TYPE:\t\t\t\t%s%n", "CL_DEVICE_TYPE_DEFAULT");
			}
			
			// CL_DEVICE_MAX_COMPUTE_UNITS
			int maxComputeUnits = getInt(device, CL_DEVICE_MAX_COMPUTE_UNITS);
			System.out.printf("CL_DEVICE_MAX_COMPUTE_UNITS:\t\t%d%n", maxComputeUnits);
			
			// CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS
			long maxWorkItemDimensions = getLong(device, CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS);
			System.out.printf("CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS:\t%d%n", maxWorkItemDimensions);
			
			// CL_DEVICE_MAX_WORK_ITEM_SIZES
			long maxWorkItemSizes[] = getSizes(device, CL_DEVICE_MAX_WORK_ITEM_SIZES, 3);
			System.out.printf("CL_DEVICE_MAX_WORK_ITEM_SIZES:\t\t%d / %d / %d %n",
					maxWorkItemSizes[0], maxWorkItemSizes[1], maxWorkItemSizes[2]);
			
			// CL_DEVICE_MAX_WORK_GROUP_SIZE
			long maxWorkGroupSize = getSize(device, CL_DEVICE_MAX_WORK_GROUP_SIZE);
			System.out.printf("CL_DEVICE_MAX_WORK_GROUP_SIZE:\t\t%d%n", maxWorkGroupSize);
			
			// CL_DEVICE_MAX_CLOCK_FREQUENCY
			long maxClockFrequency = getLong(device, CL_DEVICE_MAX_CLOCK_FREQUENCY);
			System.out.printf("CL_DEVICE_MAX_CLOCK_FREQUENCY:\t\t%d MHz%n", maxClockFrequency);
			
			// CL_DEVICE_ADDRESS_BITS
			int addressBits = getInt(device, CL_DEVICE_ADDRESS_BITS);
			System.out.printf("CL_DEVICE_ADDRESS_BITS:\t\t\t%d%n", addressBits);
			
			// CL_DEVICE_MAX_MEM_ALLOC_SIZE
			long maxMemAllocSize = getLong(device, CL_DEVICE_MAX_MEM_ALLOC_SIZE);
			System.out.printf("CL_DEVICE_MAX_MEM_ALLOC_SIZE:\t\t%d MByte%n", (int) (maxMemAllocSize / (1024 * 1024)));
			
			// CL_DEVICE_GLOBAL_MEM_SIZE
			long globalMemSize = getLong(device, CL_DEVICE_GLOBAL_MEM_SIZE);
			System.out.printf("CL_DEVICE_GLOBAL_MEM_SIZE:\t\t%d MByte%n", (int) (globalMemSize / (1024 * 1024)));
			
			// CL_DEVICE_ERROR_CORRECTION_SUPPORT
			int errorCorrectionSupport = getInt(device, CL_DEVICE_ERROR_CORRECTION_SUPPORT);
			System.out.printf("CL_DEVICE_ERROR_CORRECTION_SUPPORT:\t%s%n", errorCorrectionSupport != 0 ? "yes" : "no");
			
			// CL_DEVICE_LOCAL_MEM_TYPE
			int localMemType = getInt(device, CL_DEVICE_LOCAL_MEM_TYPE);
			System.out.printf("CL_DEVICE_LOCAL_MEM_TYPE:\t\t%s%n", localMemType == 1 ? "local" : "global");
			
			// CL_DEVICE_LOCAL_MEM_SIZE
			long localMemSize = getLong(device, CL_DEVICE_LOCAL_MEM_SIZE);
			System.out.printf("CL_DEVICE_LOCAL_MEM_SIZE:\t\t%d KByte%n", (int) (localMemSize / 1024));
			
			// CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE
			long maxConstantBufferSize = getLong(device, CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE);
			System.out.printf("CL_DEVICE_MAX_CONSTANT_BUFFER_SIZE:\t%d KByte%n", (int) (maxConstantBufferSize / 1024));
			
			// CL_DEVICE_QUEUE_PROPERTIES
			long queueProperties = getLong(device, CL_DEVICE_QUEUE_PROPERTIES);
			if ((queueProperties & CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE) != 0)
			{
				System.out.printf("CL_DEVICE_QUEUE_PROPERTIES:\t\t%s%n", "CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE");
			}
			if ((queueProperties & CL_QUEUE_PROFILING_ENABLE) != 0)
			{
				System.out.printf("CL_DEVICE_QUEUE_PROPERTIES:\t\t%s%n", "CL_QUEUE_PROFILING_ENABLE");
			}
			
			// CL_DEVICE_IMAGE_SUPPORT
			int imageSupport = getInt(device, CL_DEVICE_IMAGE_SUPPORT);
			System.out.printf("CL_DEVICE_IMAGE_SUPPORT:\t\t%d%n", imageSupport);
			
			// CL_DEVICE_MAX_READ_IMAGE_ARGS
			int maxReadImageArgs = getInt(device, CL_DEVICE_MAX_READ_IMAGE_ARGS);
			System.out.printf("CL_DEVICE_MAX_READ_IMAGE_ARGS:\t\t%d%n", maxReadImageArgs);
			
			// CL_DEVICE_MAX_WRITE_IMAGE_ARGS
			int maxWriteImageArgs = getInt(device, CL_DEVICE_MAX_WRITE_IMAGE_ARGS);
			System.out.printf("CL_DEVICE_MAX_WRITE_IMAGE_ARGS:\t\t%d%n", maxWriteImageArgs);
			
			// CL_DEVICE_SINGLE_FP_CONFIG
			long singleFpConfig = getLong(device, CL_DEVICE_SINGLE_FP_CONFIG);
			System.out.printf("CL_DEVICE_SINGLE_FP_CONFIG:\t\t%s%n",
					stringFor_cl_device_fp_config(singleFpConfig));
			
			// CL_DEVICE_IMAGE2D_MAX_WIDTH
			long image2dMaxWidth = getSize(device, CL_DEVICE_IMAGE2D_MAX_WIDTH);
			System.out.printf("CL_DEVICE_2D_MAX_WIDTH\t\t\t%d%n", image2dMaxWidth);
			
			// CL_DEVICE_IMAGE2D_MAX_HEIGHT
			long image2dMaxHeight = getSize(device, CL_DEVICE_IMAGE2D_MAX_HEIGHT);
			System.out.printf("CL_DEVICE_2D_MAX_HEIGHT\t\t\t%d%n", image2dMaxHeight);
			
			// CL_DEVICE_IMAGE3D_MAX_WIDTH
			long image3dMaxWidth = getSize(device, CL_DEVICE_IMAGE3D_MAX_WIDTH);
			System.out.printf("CL_DEVICE_3D_MAX_WIDTH\t\t\t%d%n", image3dMaxWidth);
			
			// CL_DEVICE_IMAGE3D_MAX_HEIGHT
			long image3dMaxHeight = getSize(device, CL_DEVICE_IMAGE3D_MAX_HEIGHT);
			System.out.printf("CL_DEVICE_3D_MAX_HEIGHT\t\t\t%d%n", image3dMaxHeight);
			
			// CL_DEVICE_IMAGE3D_MAX_DEPTH
			long image3dMaxDepth = getSize(device, CL_DEVICE_IMAGE3D_MAX_DEPTH);
			System.out.printf("CL_DEVICE_3D_MAX_DEPTH\t\t\t%d%n", image3dMaxDepth);
			
			// CL_DEVICE_PREFERRED_VECTOR_WIDTH_<type>
			System.out.printf("CL_DEVICE_PREFERRED_VECTOR_WIDTH_<t>\t");
			int preferredVectorWidthChar = getInt(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_CHAR);
			int preferredVectorWidthShort = getInt(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_SHORT);
			int preferredVectorWidthInt = getInt(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_INT);
			int preferredVectorWidthLong = getInt(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_LONG);
			int preferredVectorWidthFloat = getInt(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_FLOAT);
			int preferredVectorWidthDouble = getInt(device, CL_DEVICE_PREFERRED_VECTOR_WIDTH_DOUBLE);
			System.out.printf("CHAR %d, SHORT %d, INT %d, LONG %d, FLOAT %d, DOUBLE %d%n%n%n",
					preferredVectorWidthChar, preferredVectorWidthShort,
					preferredVectorWidthInt, preferredVectorWidthLong,
					preferredVectorWidthFloat, preferredVectorWidthDouble);
		}
	}
	
	
	/**
	 * Returns the value of the device info parameter with the given name
	 * 
	 * @param device The device
	 * @param paramName The parameter name
	 * @return The value
	 */
	public static int getInt(final cl_device_id device, final int paramName)
	{
		return getInts(device, paramName, 1)[0];
	}
	
	
	/**
	 * Returns the values of the device info parameter with the given name
	 * 
	 * @param device The device
	 * @param paramName The parameter name
	 * @param numValues The number of values
	 * @return The value
	 */
	public static int[] getInts(final cl_device_id device, final int paramName, final int numValues)
	{
		int values[] = new int[numValues];
		clGetDeviceInfo(device, paramName, (long) Sizeof.cl_int * numValues, Pointer.to(values), null);
		return values;
	}
	
	
	/**
	 * Returns the value of the device info parameter with the given name
	 * 
	 * @param device The device
	 * @param paramName The parameter name
	 * @return The value
	 */
	public static long getLong(final cl_device_id device, final int paramName)
	{
		return getLongs(device, paramName, 1)[0];
	}
	
	
	/**
	 * Returns the values of the device info parameter with the given name
	 * 
	 * @param device The device
	 * @param paramName The parameter name
	 * @param numValues The number of values
	 * @return The value
	 */
	public static long[] getLongs(final cl_device_id device, final int paramName, final int numValues)
	{
		long values[] = new long[numValues];
		clGetDeviceInfo(device, paramName, (long) Sizeof.cl_long * numValues, Pointer.to(values), null);
		return values;
	}
	
	
	/**
	 * Returns the value of the device info parameter with the given name
	 * 
	 * @param device The device
	 * @param paramName The parameter name
	 * @return The value
	 */
	public static String getString(final cl_device_id device, final int paramName)
	{
		// Obtain the length of the string that will be queried
		long size[] = new long[1];
		clGetDeviceInfo(device, paramName, 0, null, size);
		
		// Create a buffer of the appropriate size and fill it with the info
		byte buffer[] = new byte[(int) size[0]];
		clGetDeviceInfo(device, paramName, buffer.length, Pointer.to(buffer), null);
		
		// Create a string from the buffer (excluding the trailing \0 byte)
		return new String(buffer, 0, buffer.length - 1, Charset.forName("UTF-8"));
	}
	
	
	/**
	 * Returns the value of the platform info parameter with the given name
	 * 
	 * @param platform The platform
	 * @param paramName The parameter name
	 * @return The value
	 */
	public static String getString(final cl_platform_id platform, final int paramName)
	{
		// Obtain the length of the string that will be queried
		long size[] = new long[1];
		clGetPlatformInfo(platform, paramName, 0, null, size);
		
		// Create a buffer of the appropriate size and fill it with the info
		byte buffer[] = new byte[(int) size[0]];
		clGetPlatformInfo(platform, paramName, buffer.length, Pointer.to(buffer), null);
		
		// Create a string from the buffer (excluding the trailing \0 byte)
		return new String(buffer, 0, buffer.length - 1, Charset.forName("UTF-8"));
	}
	
	
	/**
	 * Returns the value of the device info parameter with the given name
	 * 
	 * @param device The device
	 * @param paramName The parameter name
	 * @return The value
	 */
	public static long getSize(final cl_device_id device, final int paramName)
	{
		return getSizes(device, paramName, 1)[0];
	}
	
	
	/**
	 * Returns the values of the device info parameter with the given name
	 * 
	 * @param device The device
	 * @param paramName The parameter name
	 * @param numValues The number of values
	 * @return The value
	 */
	public static long[] getSizes(final cl_device_id device, final int paramName, final int numValues)
	{
		// The size of the returned data has to depend on
		// the size of a size_t, which is handled here
		ByteBuffer buffer = ByteBuffer.allocate(
				numValues * Sizeof.size_t).order(ByteOrder.nativeOrder());
		clGetDeviceInfo(device, paramName, (long) Sizeof.size_t * numValues,
				Pointer.to(buffer), null);
		long values[] = new long[numValues];
		if (Sizeof.size_t == 4)
		{
			for (int i = 0; i < numValues; i++)
			{
				values[i] = buffer.getInt(i * Sizeof.size_t);
			}
		}
		else
		{
			for (int i = 0; i < numValues; i++)
			{
				values[i] = buffer.getLong(i * Sizeof.size_t);
			}
		}
		return values;
	}
	
}
